{$I defines.pas}
unit CETable;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  RAPI, SyncObjs, buttons, extctrls, StdCtrls, DB, dbctrls, dbtables;

const
  CM_CONNECTED = WM_USER+111;
  CE_DBNAMELENGTH = 70;
  CE_MAXDBPROPCOUNT = 120;

  PROCESSOR_ARCHITECTURE_MIPS    = 1;
  PROCESSOR_ARCHITECTURE_SHx     = 4;

type

  TCEPropType  = (ceptBlob,     // A CEBLOB structure.
                  ceptFileTime, // A FILETIME structure.
                  ceptSmallInt, // A 16-bit signed integer.
                  ceptInteger,  // A 32-bit signed integer.
                  ceptLPWSTR,   // A null-terminated string.
                  ceptWORD,     // A 16-bit unsigned integer.
                  ceptDWORD);   // A 32-bit unsigned integer.

  TDatabaseEnumEvent = procedure (Sender: TObject; Info: CEDBASEINFO; userData: TObject) of object;

  TCEInitThread = class;

  TCESession = class(TComponent)
  private
    FOnConnection: TNotifyEvent;
    FOnEnumDatabases: TDatabaseEnumEvent;
    FbConnected: boolean;
    CEInitThread: TCEInitThread;
    { Private declarations }
  protected
    { Protected declarations }
    hCEStopWaitEvent: THANDLE;
    ri: RAPIINIT;
    ahWait: array[0..1] of THANDLE;
    hParentWnd: HWND;
    lWaitingForConnection: LONG;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Connect (bSynh: boolean): boolean;
    procedure Disconnect;
    function IsConnected: boolean;
    function IsWaitingForConnection: boolean;
    procedure StopWaiting;
    procedure UploadRAPIExtensionDLL(DLLFilenameOnCE, DLLFilenameOnPC_MIPS, DLLFilenameOnPC_SH3: String);
    procedure EnumerateDatabases(userData: TObject);
  published
    { Published declarations }
    property OnConnection: TNotifyEvent read FOnConnection write FOnConnection;
    property OnEnumDatabases: TDatabaseEnumEvent read FOnEnumDatabases write FOnEnumDatabases;
  end;

  TCEInitThread = class(TThread)
  private
    CESession: TCESession;
  protected
    procedure Execute; override;
  end;

  TCETable = class;

  TCEProp = class(TComponent)
  private
    { Private declarations }
    FPropType: TCEPropType;
    FLinkTo: string;
    FPropName: string;
    FUserDefID: DWORD;
    FIndexed: boolean;
    procedure SetValues(AName: string; ALinkTo: string; ADataType: TCEPropType; AUserDefID: DWORD);
    procedure Changed;
    procedure SetUserDefID(const Value: DWORD);
    procedure SetPropType(const Value: TCEPropType);
    procedure SetPropName(const Value: string);
  protected
    { Protected declarations }
    function ReadAsInteger: integer;
    function ReadAsDateTime: TDateTime;
    function ReadAsString: string;
    function ReadAsBoolean: boolean;
    procedure WriteAsInteger (const Value: integer);
    procedure WriteAsDateTime (const Value: TDateTime);
    procedure WriteAsString (const Value: string);
    procedure WriteAsBoolean (const Value: boolean);
    procedure CheckForRecordConsistency;
  public
    { Public declarations }
    PropID: CEOID;
    ParentCETable: TCETable;
    somePointer: PWideChar;
    function PropIndex: DWORD;
    procedure AssignProp(Value: TCEProp);
    function HasParent: Boolean;  override;
    function GetParentComponent: TComponent; override;
    procedure SetParentComponent(Value: TComponent); override;
    procedure ReadState(Reader: TReader); override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IsNull: boolean;
    function GetAsBLOB (var size: integer): PChar;
    property AsInteger: integer read ReadAsInteger write WriteAsInteger;
    property AsDateTime: TDateTime read ReadAsDateTime write WriteAsDateTime;
    property AsString: string read ReadAsString write WriteAsString;
    property AsBoolean: boolean read ReadAsBoolean write WriteAsBoolean;
  published
    { Published declarations }
    property LinkTo: string read FLinkTo write FLinkTo;
    property PropName: string read FPropName write SetPropName;
    property PropType: TCEPropType read FPropType write SetPropType;
    property UserDefID: DWORD read FUserDefID write SetUserDefID;
    property Indexed: boolean read FIndexed write FIndexed;
  end;

  TCETable = class(TComponent)
  private
    FPropsList: TList;
    FTableName: string;
    FEOF, FBOF: boolean;
    FOEMConvert: boolean;
    FOnEOF: TNotifyEvent;
    FOnBOF: TNotifyEvent;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnChange: TNotifyEvent;
    dbHandle: THANDLE;
    DBID: CEOID;
    FDataSource: TDataSource;
    FOnRecordRead: TNotifyEvent;
    FOnRecordWrite: TNotifyEvent;
    function GetIndexInfo: integer;
    procedure SetDataSource(Value: TDataSource);
    function GetDataSource: TDataSource;
    procedure SetEOF(const Value: boolean);
    function GetEOF: boolean;
    procedure SetBOF(const Value: boolean);
    function GetBOF: boolean;
    { Private declarations }
  protected
    { Protected declarations }
    aCEIndexes: array [0..3] of SORTORDERSPEC;
    aCEProps: array [0..CE_MAXDBPROPCOUNT-1] of CEPROPID;
    aCEPropVal: array [0..CE_MAXDBPROPCOUNT-1] of CEPROPVAL;
    Function AppendInt: boolean;
    function WriteRecordInt (oidRecord: integer):boolean;
    procedure ClearRecord;
    property IntEOF: boolean read GetEOF write SetEOF;
    property IntBOF: boolean read GetBOF write SetBOF;
  public
    { Public declarations }
    curRecord: PChar;
    curRecordSize: DWORD;
    curRecordCEOID: DWORD;
    PropCountWrite: DWORD;
    dwRecNo: DWORD;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    procedure Next;
    procedure Prior;
    procedure MoveBy (recNum: integer);
    procedure Last;
    procedure First;
    function ReadRecord: boolean;
    function FindProp (idProp: DWORD): TCEProp;
    function PropByName (prop_name: string): TCEProp;
    function PropsCount: integer;
    procedure Post;
    procedure Append;
    procedure Edit;
    procedure Cancel;
    Function Delete: boolean;
    procedure CreateDatabase;
    Function DeleteDatabase: boolean;
    Function RecordCount: integer;
    function AddProp(AName: string; ALinkTo: string; ADataType: TCEPropType): Longint; overload;
    function AddProp(AName: string; ALinkTo: string; ADataType: TCEPropType; AUserDefID: DWORD): Longint; overload;
    procedure DeleteProp(Index: integer);
    procedure RemoveProp(Item: TCEProp);
    function InsertProp(Item: TCEProp):integer;
    procedure GetChildren(Proc: TGetChildProc;
      Root: TComponent); override;
    procedure UploadTable;
    procedure DownloadTable;
    property EOF: boolean read FEOF write FEOF;
    property BOF: boolean read FBOF write FBOF;
    procedure PopulatePropsByRecordContent;
  published
    { Published declarations }
    property TableName: string read FTableName write FTableName;
    property PropsList: TList read FPropsList;
    property OEMConvert: boolean read FOEMConvert write FOEMConvert;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property OnRecordRead: TNotifyEvent read FOnRecordRead write FOnRecordRead;
    property OnRecordWrite: TNotifyEvent read FOnRecordWrite write FOnRecordWrite;
    property OnEOFChanged: TNotifyEvent read FOnEOF write FOnEOF;
    property OnBOFChanged: TNotifyEvent read FOnBOF write FOnBOF;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  TformConnectDialog = class(TForm)
    BitBtn1: TBitBtn;
    Image1: TImage;
    Label1: TLabel;
    procedure BitBtn1Click(Sender: TObject);
  private
    CESession: TCESession;
    { Private declarations }
  public
    { Public declarations }
    procedure WMUserMsg(var Message: TMessage); message CM_CONNECTED;
  end;

  TCESort = (CEDB_SORT_DESCENDING,      {The sort is done in descending order. By default, the sort is done in ascending order.}
             CEDB_SORT_CASEINSENSITIVE, {The sort opertion is case sensitive. This value is valid only for strings.}
             CEDB_SORT_UNKNOWNFIRST,    {Records that do not contain this property are placed before all the other records. By default,
                                         such records are placed after all other records.}
             CEDB_SORT_GENERICORDER);
Var
   formConnectDialog: TformConnectDialog;

function CEFormatError (errCode: DWORD):string;
function GetCETypeViaStr (sPropType: string): WORD;
function GetCETypeViaEnumID (CEType: TCEPropType): WORD;
function GetEnumIDViaCEType (CEType: Word): TCEPropType;

implementation
